
import os
import fish_feats.Utils as ut
import fish_feats.FishWidgets as fwid
from fish_feats.NapaMix import Separation
import numpy as np
from qtpy.QtWidgets import QVBoxLayout, QWidget 

class MainCells( QWidget ):
    """ Main interface for cell (junctions) segmentation """

    def __init__( self, ffeats ):
        """ Interface to choose loading or proj and seg """
        super().__init__()
        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg
        self.ffeats = ffeats
        self.proj = None
    
        methods = ["", "Do projection and segmentation"]
        ind = 0
        self.projname = self.mig.build_filename( "_junction_projection.tif")
        self.cellsname = self.mig.build_filename( "_cells2D.tif")
        msg = ""
        if os.path.exists( self.projname ):
            msg = "Found projection file"
        if os.path.exists( self.cellsname ):
            methods.append( "Load previous files" )
            ind = 1
            msg += "\nFound cell file"
        msg += "\nChoose load to use those file(s)"

        if len(methods) == 2:
            ## there is nothing to load, go directly to projection and segmentation
            self.proj = Projection( self.ffeats )
            return

        layout = QVBoxLayout()
        ## Add message of found files
        info_msg = fwid.add_label( ""+msg, descr="Show info if found default files that can be loaded or not" )
        layout.addWidget( info_msg )
        ## Add choice of action
        do_line, self.methodChoice = fwid.list_line( "", descr="Choose to do projection and segmentation or load previous files", func=self.go_segmentation )
        self.methodChoice.addItems( methods )
        layout.addLayout( do_line )
        self.setLayout( layout )

    def go_segmentation( self ):
        """ Start the process according to selected method """
        method = self.methodChoice.currentText()
        if method == "Do projection and segmentation":
            ## go to projection plugin
            proj = Projection( self.ffeats )
            self.viewer.window.add_dock_widget( proj, name="JunctionProjection2D" )
        elif method == "Load previous files":
            ## load and show the projection
            if os.path.exists( self.projname ):
                ut.remove_layer( self.viewer, "2DJunctions" )
                roijunc = self.mig.load_image( self.projname )
                self.viewer.add_image( roijunc, name="2DJunctions", scale=(self.mig.scaleXY, self.mig.scaleXY), blending="additive" )
            ## load the cells and edit them
            self.mig.load_segmentation( self.cellsname )
            #get_cells = GetCells( self.ffeats )
            self.ffeats.correction_junctions()
            self.close()

class Projection( QWidget ):
    """ Get the 2D projection (local) of the junctions image """

    def __init__( self, ffeats ):
        """ Interface to handle local projection """
        super().__init__()
        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg
        self.ffeats = ffeats
        self.projection_filename = None
        self.projection_filename = self.mig.junction_projection_filename( ifexist=True )

        print("******** Junction staining 2D projection for segmentation ******")
        help_text = ut.help_shortcut( "projection" )
        ut.showOverlayText( self.viewer, help_text )

        layout = QVBoxLayout()

        ## Path to projection image file
        self.file_group, filename_layout = fwid.group_layout( "Load projection file", "Options to load projection image file" )
        load_btn = fwid.add_button( "Load default", self.load_projection_file, descr="Load the projection of junctions image", color=ut.get_color("load") )
        choose_file = fwid.add_button( "or choose file", self.get_projection_filename, descr="Choose a file with projected junctions image", color=None )
        load_line = fwid.double_widget( load_btn, choose_file )
        filename_layout.addLayout( load_line )
        self.file_group.setLayout( filename_layout )
        layout.addWidget( self.file_group )
        if (self.projection_filename is None) or ( self.projection_filename == "" ):
            load_btn.setEnabled( False )

        ## Additional options: projection parameters
        grp_line, ad_check, self.advanced = fwid.checkgroup_help( "Advanced", True, "Options for the projection calculation", help_link="Get-cells#2d-projection" )
        adv_layout = QVBoxLayout() 
        # local size
        wsize_line, self.local_size = fwid.value_line( "Local size", 40, descr="Size of the local projection in pixels" )
        ## smoothing size
        smooth_line, self.smooth_size = fwid.value_line( "Smoothing size", 3, descr="Smoothing of the local projection in pixels" )
        ## local contrast
        self.do_clahe = fwid.add_check( "Do local enhancement", False, None, descr="Apply local enhancement CLAHE to the projection image" )
        clahe_size, self.clahe_size = fwid.value_line( "CLAHE grid size", 20, descr="Grid size for the local enhancement CLAHE" )
        adv_layout.addLayout( wsize_line )
        adv_layout.addLayout( smooth_line )
        adv_layout.addWidget( self.do_clahe )
        adv_layout.addLayout( clahe_size )
        self.advanced.setLayout( adv_layout )
        layout.addLayout( grp_line )
        layout.addWidget( self.advanced )
        ad_check.setChecked( False )

        ## Launch calculation of the projection
        proj_btn = fwid.add_button( "Project now", self.do_projection, descr="Calculate the projection of junctions image", color=ut.get_color("go") )
        layout.addWidget( proj_btn )
        
        ## Save the projection to file after calculation
        self.save_proj = fwid.add_check( "Save projection", True, None, descr="Save the projection to file after calculation" )
        layout.addWidget( self.save_proj )

        ## Finish the step
        done_btn = fwid.add_button( "Projection done", self.finish_projection, descr="Finish the projection step and go to segmentation", color=ut.get_color("done") )
        layout.addWidget( done_btn )
        self.setLayout(layout)

    def do_projection( self ):
        """ Load/Calculate the projection, save and go to next step. Separate the signals if necessary before """
        ut.remove_layer( self.viewer, "2DJunctions" )
        ## separate if necessary the signals
        if self.mig.should_separate():
            ut.show_info("Junctions and nuclei staining in the same channel, separate them first")
            separation = Separation( self.viewer, self.mig, self.cfg ) 
            self.viewer.window.add_dock_widget(separation, name="Separate")
            return
        else:
            ## calculates the projection
            projxy = int( float( self.local_size.text() ) )
            smooth = int( float( self.smooth_size.text() ) )
            do_clahe = self.do_clahe.isChecked()
            clahe_grid = int( float( self.clahe_size.text() ) )
            roijunc = self.mig.prepare_segmentation_junctions( projxy=projxy, projsmooth=smooth, do_clahe=do_clahe, clahe_grid=clahe_grid )
            self.viewer.add_image( roijunc, name="2DJunctions", scale=(self.mig.scaleXY, self.mig.scaleXY), blending="additive" )

    def finish_projection( self ):
        """ Projection done, go to the next step (segmentation) """
        ## save the results if option is on
        if self.save_proj.isChecked():
            if "2DJunctions" not in self.viewer.layers:
                ut.show_warning( "Projected layer 2DJunctions not found" )
                return
            roijunc = self.viewer.layers["2DJunctions"].data
            outname = self.mig.build_filename( "_junction_projection.tif")
            self.mig.save_image( roijunc, imagename=outname )

        ut.removeOverlayText(self.viewer)
        ut.remove_layer( self.viewer, "junctionsStaining" )
        ut.remove_layer( self.viewer, "nucleiStaining" )
        ut.remove_widget( self.viewer, "Get cells" )
        ut.remove_widget( self.viewer, "JunctionProjection2D" )
        self.go_segmentation()
    
    def go_segmentation( self ):
        """ Segmentation then correction of the apical junctions """
        ut.show_info("******** Segmentation of junctions 2D projection ******")
        get_cell = GetCells( self.ffeats )
        self.viewer.window.add_dock_widget( get_cell, name="Segment cells" )

    def load_projection_file( self ):
        """ Load the projection file (default or selected) """
        ut.remove_layer( self.viewer, "2DJunctions" )
        roijunc = self.mig.load_image( self.projection_filename )
        self.viewer.add_image( roijunc, name="2DJunctions", scale=( self.mig.scaleXY, self.mig.scaleXY), blending="additive" )
        self.finish_projection()

    def get_projection_filename( self ):
        """ Open a file dialog to choose a file with projection of junctions """
        filename = fwid.file_dialog( "Choose projected junctions file", "*.tif", directory=self.mig.resdir )
        if filename is not None:
            self.projection_filename = filename
            #self.filename_label.setText(filename)
            self.load_projection_file()

class GetCells( QWidget ):
    """ Segment the cell contours from the image """

    def __init__( self, ffeats ):
        """ Interface to get the cell contours from the image """
        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg
        self.ffeats = ffeats
        self.junction_filename = None

        super().__init__()
        layout = QVBoxLayout()

        methods = [ "Epyseg", "CellPose", "Load segmented file", "Empty" ]
        defmeth = "Epyseg"
        celldiameter = 30
        self.chunksize = 1500
        self.paras = self.cfg.read_junctions()
        if self.paras is not None:
            if "chunk_size" in self.paras:
                self.chunksize = int(float(self.paras["chunk_size"]))
            if "cell_diameter" in self.paras:
                celldiameter = int(float(self.paras["cell_diameter"]))
            if "method" in self.paras:
                defmeth = self.paras["method"]
        if self.mig.junction_filename(dim=2, ifexist=True) != "":
            defmeth = "Load segmented file"
        self.junction_filename = self.mig.junction_filename(dim=2, ifexist=True)
        if self.junction_filename is None:
            self.junction_filename = self.mig.resdir

        ut.showOverlayText(self.viewer, "Choose cell junctions segmentation option", size=14)
        ut.hide_color_layers(self.viewer, self.mig)
        ut.show_layer(self.viewer, self.mig.junchan)

        ## choose the method
        meth_line, self.methodsChoice = fwid.list_line( "Method",  descr="Choose the method to segment the cell contours" )
        for meth in methods:
            self.methodsChoice.addItem( meth )
        layout.addLayout( meth_line )
        ## choose the cell diameter
        self.diam_group, diam_layout = fwid.group_layout( "", "" )
        diam_line, self.diameter = fwid.value_line( "Cell diameter", celldiameter, descr="Mean diameter of cell in pixels, used for segmentation" )
        diam_layout.addLayout( diam_line )
        self.diam_group.setLayout( diam_layout )
        layout.addWidget( self.diam_group )
        ## choose the chunk size
        #self.chunk_line, self.chunk_size = fwid.value_line( "Chunk size", chunksize, descr="Chunk size for the segmentation, used to avoid memory issues" )
        #layout.addLayout( self.chunk_line )
        ## choose loading filename
        self.file_group, filename_layout = fwid.group_layout( "", "" )
        filename_line, self.filename_label = fwid.file_line( "Choose file", self.junction_filename, dial_msg="Choose segmentation file", filetype="*.tif", descr="Choose the file containing the segmentation of the cells (labels)" )
        #filename_line, choose_filename, self.filename_label = fwid.label_button( "Choose file", self.get_junction_filename, label=self.junction_filename, descr="Choose a file with segmented junctions", color=None )
        filename_layout.addLayout( filename_line)
        self.file_group.setLayout( filename_layout )
        layout.addWidget( self.file_group )

        ## button to segment the cells
        segment_btn = fwid.add_button( "Segment cells", self.segment_cells, descr="Segment the cell contours from the image", color=ut.get_color("go") )
        layout.addWidget( segment_btn )
    
        self.methodsChoice.currentIndexChanged.connect( self.visibility )
        self.methodsChoice.setCurrentText( defmeth )
        self.visibility()
        self.setLayout( layout )

    def visibility( self ):
        """ Set the visibility of the parameters according to the method """
        self.diam_group.setVisible( self.methodsChoice.currentText() in ["CellPose"] )
        self.file_group.setVisible( self.methodsChoice.currentText() == "Load segmented file" )

    def get_junction_filename( self ):
        """ Open a file dialog to choose a file with segmented junctions """
        filename = fwid.file_dialog( "Choose segmented junctions file", "*.tif", directory=self.mig.resdir )
        if filename is not None:
            self.junction_filename = filename
            self.filename_label.setText(filename)

    def segment_cells( self ):
        """ Segment the cell contours from the image """
        ut.showOverlayText( self.viewer, """Doing segmentation of cell junctions...""", size=15 )
        self.cfg.addGroupParameter("JunctionSeg")
        method = self.methodsChoice.currentText()
        self.cfg.addParameter("JunctionSeg", "method", method)
        self.cfg.addParameter("JunctionSeg", "cell_diameter", int(self.diameter.text()))
        self.cfg.addParameter("JunctionSeg", "chunk_size", self.chunksize)
        self.cfg.write_parameterfile()
        ut.remove_layer( self.viewer, "Junctions" )
        if method == "Load segmented file":
            filename = self.filename_label.text()
            if os.path.exists( filename ):
                self.junction_filename = filename
            self.mig.load_segmentation( self.junction_filename )
            self.ffeats.correction_junctions()
            self.close()
        else:
            if "2DJunctions" in self.viewer.layers:
                roijunc = self.viewer.layers["2DJunctions"].data
                self.mig.do_segmentation_junctions( method, roijunc, int(self.diameter.text()), self.chunksize )
                self.ffeats.correction_junctions()
                self.close()
            else:
                ut.show_info("No projected junctions to segment, go to projection")
                ut.remove_widget(self.viewer, "Segment cells")
                ut.remove_widget( self.viewer, "Get cells" )
                proj = Projection( self.ffeats )
                self.viewer.window.add_dock_widget( proj, name="JunctionProjection2D" )


class EndCells( QWidget ):
    """ Handle the finishing of cell contour edition """

    def __init__( self, ffeats):
        """ Interface to save/finish the cell edition option """
        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg
        self.ffeats = ffeats
        
        super().__init__()
        layout = QVBoxLayout()

        ## show table button
        show_btn = fwid.add_button( "Show measures", self.show_measures, descr="Show the table of cell positions and areas" )
        layout.addWidget( show_btn )

        ## save current state of cells
        save_btn = fwid.add_button( "Save cells", self.save_all, descr="Save the current segmentation and cell measures", color=ut.get_color("save") )
        layout.addWidget( save_btn )

        ## options to save and do 3d position
        self.find_z_positions = fwid.add_check( "Find 3D positions", True, None, descr="Place the cell in 3D when finishing this step" )
        self.save_when_done = fwid.add_check( "Save when done", True, None, descr="Save the results and segmentation when this option is finished" )
        check_line = fwid.double_widget( self.find_z_positions, self.save_when_done )
        layout.addLayout( check_line )

        ## Final button, all done
        finish_btn = fwid.add_button( "Cells done", self.finish_junctions, descr="Finish this step, save if necessary", color=ut.get_color("done") )
        layout.addWidget( finish_btn )
    
        self.cfg.addText("'show_measures' open the table of cells measurement: area and position\n")
        self.setLayout( layout )

    def show_measures( self ):
        """ Display the measure table """
        self.mig.popFromJunctions()
        results = self.mig.measure_junctions()
        ut.show_table( results )
    
    def save_junc( self ):
        """ Save image of the labelled cells """
        filename = self.mig.junction_filename(dim=2, ifexist=False)
        self.mig.save_image( self.viewer.layers["Junctions"].data, filename, hasZ=False )
    
    def save_all( self ):
        """ Save the current image of the cells and the table results """
        self.save_junc()
        self.mig.save_results()

    def finish_junctions( self ):
        """ Finish the get cell option """
        self.mig.popFromJunctions( zpos = self.find_z_positions.isChecked() )
        if self.save_when_done.isChecked():
            self.save_all()
        self.viewer.window.remove_dock_widget("all")
        ut.removeOverlayText( self.viewer )
        ut.remove_layer( self.viewer, "Junctions" )
        ut.remove_layer( self.viewer, "2DJunctions" )
        ut.remove_layer( self.viewer, "JunctionsName" )
        self.cfg.removeTmpText()
        self.ffeats.getChoices( default_action = "Get nuclei" )


class Position3D( QWidget ):
    """ Interface to set/correct the cell 3D position """

    def __init__( self, ffeats ):
        """ GUI to handle cell 3D position """
        super().__init__()
        self.viewer = ffeats.viewer
        self.mig = ffeats.mig
        self.cfg = ffeats.cfg
    
        print("******** Cells Z position viewing/editing ******")
        header = ut.helpHeader(self.viewer, "CellContours")
        help_text = ut.help_shortcut("pos3d")
        ut.showOverlayText( self.viewer, header+help_text)
        paras = self.cfg.read_parameter_set("ZCells")
        zmapres = 200
        zmaplocsize = 300
        if paras is not None:
            if "zmap_resolution" in paras:
                zmapres = int(paras["zmap_resolution"])
            if "zmap_localsize" in paras:
                zmaplocsize = int(paras["zmap_localsize"])

        layout = QVBoxLayout()
        ## calculate map
        map_lab = fwid.add_label( "Map of cell Z position", descr="Paramters to calculate the local height map of cell positions")
        layout.addWidget( map_lab )
        ## choose resolution
        res_line, self.resolution = fwid.value_line(" Zmap resolution:", zmapres, descr="Resolution of the resulting Zmap after calculation (local neighborhood of same Z size)" )
        layout.addLayout( res_line )
        ## choose localsize 
        loc_line, self.localsize = fwid.value_line(" Zmap local size:", zmaplocsize, descr="Size of the local neighborhood to calculate the Z position" )
        layout.addLayout( loc_line )
        ## save option
        self.save_map = fwid.add_check( "Save Zmap", True, None, descr="Save the Zmap after calculation" )
        layout.addWidget( self.save_map )
        ##btn go calculate
        calc_btn = fwid.add_button( "(Re)calculate Zmap", self.calculate_zmap, descr="Calculate the Zmap with the selected parameters", color=ut.get_color("go") )
        layout.addWidget( calc_btn )

        ## add a blank space
        space = fwid.add_label( " ", descr="" )
        layout.addWidget( space )

        ### Edit map part
        ## Name of the part
        edit_lab = fwid.add_label( "Edit cell Z position", descr="Edit the Z position of a cell parameters" )
        layout.addWidget( edit_lab )
        ## cell label choice
        celllab_line, self.cell_label = fwid.value_line( "Cell label:", 0, descr="Label of the cell to edit" )
        layout.addLayout( celllab_line )
        ## new position
        position_line, self.new_position = fwid.value_line( "New Z position:", 0, descr="New Z position to place the cell" )
        layout.addLayout( position_line )
        ## btn go for the cell
        poscell_btn = fwid.add_button( "Update selected cell position", self.update_cell_zpos, descr="Update the cell to the new Z position", color=ut.get_color("go") )
        layout.addWidget( poscell_btn )

        ## Space
        space2 = fwid.add_label( " ", descr="" )
        layout.addWidget( space2 )

        ## Save all button
        saveall_btn = fwid.add_button( "Save updated cells", self.save_all, descr="Save the current cell Z positions", color=ut.get_color("save") )
        ## finish button
        end_btn = fwid.add_button( "3D Position done", self.end_position, descr="Finish this step, save and close the panel", color=ut.get_color("done") )
        btn_line = fwid.double_button( saveall_btn, end_btn )
        layout.addLayout( btn_line ) 
        self.setLayout( layout )

        self.drawCells3D()
    
    def calculate_zmap( self ):
        """ recalculate the zmap and update cells positions """
        step_size = int( float(self.resolution.text() ) )
        window_size = int( float(self.localsize.text() ) )
        self.mig.updateCellsZPos( step_size=step_size, window_size=window_size, save=self.save_map.isChecked() )
        self.drawCells3D()
        ut.show_info("Cell Z positions updated")

    def drawCells3D(self):
        """ Draw the cells in 3D """
        ready = self.mig.cellsHaveZPos()
        if not ready:
            ## more than half the cells don't have Z position, so recompute it
            ut.show_info("Many cells don't have Z position yet, computing it")
            self.save_map.setChecked( False )
            self.calculate_zmap()
        cells3D = self.mig.getJunctionsImage3D()
        ut.remove_layer(self.viewer, "CellContours")
        layer = self.viewer.add_labels( cells3D, name="CellContours", blending="additive", scale=(self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY) )

        @layer.mouse_drag_callbacks.append
        def clicks_label(layer, event):
            if event.type == "mouse_press":
                if len(event.modifiers) == 0:
                    if event.button == 2:
                        # right-click, select the label value
                        label = layer.get_value(position=event.position, view_direction = event.view_direction, dims_displayed=event.dims_displayed, world=True)
                        if label > 0:
                            self.cell_label.setText( str(label) )
                    return
                if "Control" in event.modifiers:
                    if event.button == 1:
                        ## Control left-click, set the z position
                        zpos = self.viewer.dims.current_step[0]
                        self.new_position.setText( str(zpos) )
                        self.update_cell_zpos()
                        return

    def update_cell_zpos(self):
        """ Update current cell to current z position """
        cell = int( self.cell_label.text() )
        zpos = int( self.new_position.text() )
        if (zpos >= 0) and (zpos < self.mig.get_image_shape(in2d=False)[0]):
            img = self.viewer.layers["CellContours"].data
            self.mig.updateCellZPos(cell, zpos, img)
            #self.viewer.layers["CellContours"].data = img
            self.viewer.layers["CellContours"].refresh()
            self.cell_label.setText( "0" )

    def save_all( self ):
        """ Save all updated cell positions """
        self.mig.save_results()
        self.cfg.addGroupParameter( "ZCells" )
        self.cfg.addParameter( "ZCells", "zmap_resolution", int( float(self.resolution.text()) ) )
        self.cfg.addParameter( "ZCells", "zmap_window_size", int( float(self.localsize.text()) ) )
        self.cfg.write_parameterfile()

    def end_position( self ):
        """ Finish and save this step """
        self.save_all()
        ut.remove_widget( self.viewer, "Cells in 3D" )
        ut.remove_layer( self.viewer, "CellContours" )
        ut.removeOverlayText( self.viewer )

"""
def preprocJunctions2D(imgjun):
    "" Preprocessing the projection (filters, denoising)""

    saveimg = np.copy(imgjun)
    if "2DJunctions" not in viewer.layers:
        viewer.add_image( imgjun, name="2DJunctions", blending="additive", scale=(mig.scaleXY, mig.scaleXY), colormap="red" )

    def update_parameters():
        removebg_parameters(preprocess.remove_background.value)
        tophat_parameters(preprocess.tophat_filter.value)
        n2v_parameters(preprocess.noise2void.value)
    
    def reset_junc():
        ut.remove_layer(viewer,"2DJunctions")
        imgjun = saveimg
        viewer.add_image( imgjun, name="2DJunctions", blending="additive", scale=(mig.scaleXY, mig.scaleXY), colormap="red" )
    
    def n2v_parameters(booly):
        preprocess.denoising_done.visible = booly

    def end_denoising():
        if "Denoised" in viewer.layers:
            ut.remove_widget(viewer, "Dock widget 1")
            ut.remove_layer(viewer, "2DJunctions")
            viewer.layers["Denoised"].name = "2DJunctions"
        viewer.layers["2DJunctions"].refresh()

    def removebg_parameters(booly):
        preprocess.remove_background_radius.visible = booly
    
    def tophat_parameters(booly):
        preprocess.tophat_filter_radius.visible = booly

    @magicgui(call_button="Preprocess", 
            reset_junction_staining={"widget_type":"PushButton", "value": False, "name": "reset_junction_staining"}, 
            denoising_done={"widget_type":"PushButton", "value": False, "name": "denoising_done"}, )
    def preprocess(
            remove_background=False,
            remove_background_radius = 50,
            tophat_filter=False,
            tophat_filter_radius = 5,
            noise2void = False,
            reset_junction_staining=False,
            denoising_done = False,
            ):
        cfg.addText("Preprocess 2D junction staining")
        
        imgjun = viewer.layers["2DJunctions"].data
        if remove_background:
            #cfg.addTextParameter("Preprocess", "remove_background_radius", remove_background_radius)
            imgjun = mig.preprocess_junction2D_removebg( imgjun, remove_background_radius )
            ut.show_info("background removed")
        
        if tophat_filter:
            #cfg.addTextParameter("Preprocess", "tophat_filter_radius", tophat_filter_radius)
            imgjun = mig.preprocess_junction2D_tophat( imgjun, tophat_filter_radius )
            ut.show_info("Tophat filter applied")
        
        if "2DJunctions" in viewer.layers:
            viewer.layers["2DJunctions"].data = imgjun
            viewer.layers["junctionsStaining"].refresh()
        
        if noise2void:
            #mig.prepare_junctions()
            from napari_n2v import PredictWidgetWrapper
            viewer.window.add_dock_widget(PredictWidgetWrapper(viewer))

        if not "2DJunctions" in viewer.layers:
            viewer.add_image( imgjun, name="2DJunctions", blending="additive", scale=(mig.scaleXY, mig.scaleXY), colormap="blue" )
        viewer.layers["2DJunctions"].data = imgjun
        viewer.layers["2DJunctions"].refresh()
    
    removebg_parameters(False)
    tophat_parameters(False)
    n2v_parameters(False)
    preprocess.remove_background.changed.connect(update_parameters)
    preprocess.tophat_filter.changed.connect(update_parameters)
    preprocess.noise2void.changed.connect(update_parameters)
    preprocess.reset_junction_staining.clicked.connect(reset_junc)
    preprocess.denoising_done.clicked.connect(end_denoising)
    ut.hide_color_layers(viewer, mig)
    if "junctionsStaining" in viewer.layers:
        viewer.layers["junctionsStaining"].visible = True
    viewer.window.add_dock_widget(preprocess, name="Preprocess2D")
"""


import numpy as np
import napari
from napari.utils.notifications import show_info
#from magicgui import magicgui
from magicgui.widgets import Table
import fish_feats.MainImage as mi
import fish_feats.Utils as ut
import fish_feats.FishWidgets as fwid
import pathlib
from qtpy.QtWidgets import QPushButton, QHBoxLayout, QVBoxLayout, QWidget, QGroupBox, QLineEdit, QComboBox, QLabel, QSpinBox, QCheckBox, QTabWidget, QFileDialog, QTableWidget, QTableWidgetItem, QGridLayout
from qtpy.QtCore import Qt
import time

""" 
Widget to directly use the ClassifyCells option without going again through all the pipeline.
Needs the segmented cells file.
"""

def get_image():
    """ Load and initialize the desired file """
    mig = mi.MainImage( talkative=True )
    viewer = napari.current_viewer()
    viewer.title = "Fish&Feats"
    viewer.scale_bar.visible = True
    
    @viewer.bind_key('h', overwrite=True)
    def show_help(layer):
        ut.showHideOverlayText(viewer)
    
    @viewer.bind_key('F1', overwrite=True)
    def show_layer(viewer):
        show_hide( 0 )
        
    @viewer.bind_key('F2', overwrite=True)
    def show_layer(viewer):
        show_hide( 1 )
        
    @viewer.bind_key('F3', overwrite=True)
    def show_layer(viewer):
        show_hide( 2 )
        
    @viewer.bind_key('F4', overwrite=True)
    def show_layer(viewer):
        show_hide( 3 )
        
    @viewer.bind_key('F5', overwrite=True)
    def show_layer(viewer):
        show_hide( 4 )
        
    @viewer.bind_key('F6', overwrite=True)
    def show_layer(viewer):
        show_hide( 5 )
    
    @viewer.bind_key('F7', overwrite=True)
    def show_layer(viewer):
        show_hide( 6 )
    
    @viewer.bind_key('F8', overwrite=True)
    def show_layer(viewer):
        show_hide( 7 )
    
    def show_hide( intlayer ):
        """ Show/hide the ith-layer """
        if intlayer < len( viewer.layers ):
            viewer.layers[intlayer].visible = not viewer.layers[intlayer].visible

    
    filename = ut.dialog_filename()
    if filename is None:
        print("No file selected")
        return
    
    mig.open_image( filename=filename )
    ut.update_history(mig.imagedir)
    
    for channel in range(mig.nbchannels):
        cmap = ut.colormapname(channel)
        img = mig.get_channel(channel)
        cview = viewer.add_image( img, name="originalChannel"+str(channel), blending="additive", scale=(mig.scaleZ, mig.scaleXY, mig.scaleXY), colormap=cmap )
        dint = np.max(img)-np.min(img)
        cview.contrast_limits=(np.min(img), np.max(img)-0.75*dint)
    viewer.axes.visible = True

    getscales = GetScales( viewer, mig, classify_cells )
    wid = viewer.window.add_dock_widget(getscales, name="Scale")
    return wid


################ attribute cells
def classify_cells(mig, viewer):
    """ Load main interface of classify cells option """ 
    print("************ Classify cells by user-defined category **************")

    def showHelp():
        """ Open wiki documentation page """
        ut.show_documentation_page("Classify-cells")
    
    def saveTable():
        """ Save the table of features, in the results file """
        ## update the table with edited feature
        featTable.set_table()
        mig.save_results()

    def endMain():
        """ Close all widgets and save results """
        mig.save_results()
        ut.remove_layer(viewer, "Cells")
        for feat in mig.getFeaturesList():
            ut.remove_layer(viewer, "ProjC"+feat)
            ut.remove_layer(viewer, feat+"Cells")
        wids = []
        for wid in ut.list_widgets(viewer):
            if wid.startswith("Feat_"):
                wids.append(wid)
            if wid.startswith("Edit Feat_"):
                wids.append(wid)
        for wid in wids:
            ut.remove_widget(viewer, wid)
        ut.remove_widget(viewer, "Main")
        ut.removeOverlayText(viewer)

    ut.remove_layer(viewer, "Cells")
    labimg = viewer.add_labels( mig.pop.imgcell, name="Cells", scale=(mig.scaleXY, mig.scaleXY), opacity=0.4, blending="additive" )
    labimg.contour = 1

    if not ut.has_widget( viewer, "Classify cells"):
        main_widget = QWidget()
        layout = QVBoxLayout()
        main_widget.setLayout(layout)
        
        btn_line = QHBoxLayout()
        ## save table option
        save_table = fwid.add_button( "Update/Save table", saveTable, descr="Update the displayed table with the modified feature and save it in the results file", color=ut.get_color("save") )
        #save_table = QPushButton("Update/Save table")
        btn_line.addWidget(save_table)
        
        helpb = QPushButton("Help")
        btn_line.addWidget(helpb)
        helpb.setStyleSheet( 'QPushButton {background-color: '+ut.get_color("help")+' } ')
        helpb.clicked.connect(showHelp)
        
        ## finish classify option
        end_main = QPushButton("Stop and Save")
        btn_line.addWidget(end_main)
        end_main.clicked.connect(endMain)
        end_main.setStyleSheet( 'QPushButton {background-color: '+ut.get_color("done")+'} ')
        layout.addLayout(btn_line)
        
        tabs = QTabWidget()
        tabs.setObjectName("Classify cells")
        layout.addWidget(tabs)

        ## Fetures table        
        featTable = FeaturesTable(viewer, mig)
        tabs.addTab( featTable, "Feature table" )
        add_feat = AddFeature(viewer, mig, labimg, featTable)
        tabs.addTab( add_feat, "Add/Edit a feature" )
        

        viewer.window.add_dock_widget( main_widget, name="Main" )
        text = "Classify manually cells into categories \n"
        text += ut.help_shortcut("classify")
        ut.showOverlayText(viewer, text)
        
        

class HandleTable(QWidget):
    """ Handle the table of all the features """
    def __init__(self, napari_viewer, mig, featTable):
        super().__init__()
        self.viewer = napari_viewer
        self.mig = mig
        self.table = featTable
        layout = QVBoxLayout()

        self.setLayout(layout)

class AddFeature(QWidget):
    """ Options to add a feature to measure """

    def __init__(self, napari_viewer, mig, labimg, table):
        super().__init__()
        self.viewer = napari_viewer
        self.mig = mig
        self.labimg = labimg
        self.table = table
        self.nb_values = {}
        
        layout = QVBoxLayout()
        
        feature_line = QHBoxLayout()
        feat_btn = QPushButton( "Do feature" )
        feature_line.addWidget( feat_btn )
        feat_btn.setToolTip( "Click to create/edit the selected feature" )
        feat_btn.setStatusTip( "Click to create/edit the selected feature" )
        ## Show editable list of features
        self.features_list = QComboBox()
        if self.mig.getFeaturesList() is not None:
            for feat in mig.getFeaturesList():
                self.features_list.addItem(feat)
        feature_line.addWidget( self.features_list )
        self.features_list.setEditable(True)
        self.features_list.setToolTip("Select a feature to edit or write the name of a new feature")
        feat_btn.clicked.connect( self.do_feature )
        layout.addLayout( feature_line )        
            
        self.setLayout(layout)
        self.creator = CreateFeature( self )
       

    def do_feature( self ):
        """ Displays the correct GUI depending if creating new feature or editing one """
        if self.features_list.currentText() == "":
            ut.show_warning("Please select a feature to edit or write the name of a new feature")
            return
        if (self.features_list.currentText() in self.mig.getFeaturesList()) or ("Feat_"+self.features_list.currentText() in self.mig.getFeaturesList()):
            self.go_edit_feature()
        else:
            ## creating a new feature
            self.creator.set_name()
            self.creator.show()
    
    def create_editfeature(self):
        """ GUI of edit feature option """
        geditfeat_layout = QVBoxLayout()
        
        loadfile = QHBoxLayout()
        self.loadTableFile = QLineEdit()
        self.loadTableFile.setText(self.mig.features_filename(ifexist=True))
        loadfile.addWidget(self.loadTableFile)
        select_file = QPushButton("Load feature(s) from file", parent=self)
        loadfile.addWidget(select_file)
        select_file.clicked.connect(self.selectFile)
        geditfeat_layout.addLayout(loadfile)
        
        self.choose_feat = QComboBox()
        geditfeat_layout.addWidget(self.choose_feat)
        features_list = self.table.get_features_list()
        for feat in features_list:
            if (feat != "CellLabel") and (feat != "CellId"):
                self.choose_feat.addItem(feat)
        
        edit_feat = QPushButton("Edit selected feature", parent=self)
        geditfeat_layout.addWidget(edit_feat)
        edit_feat.clicked.connect(self.go_edit_feature)
        
        self.gEditFeat.setLayout(geditfeat_layout)
        self.gEditFeat.setStyleSheet("QGroupBox { color: #d6d4df; background-color: #363650; }")

    def update_editfeature(self):
        """ Update the list of loaded features """
        self.choose_feat.clear()
        features_list = self.table.get_features_list()
        for feat in features_list:
            if (feat != "CellLabel") and (feat != "CellId"):
                self.choose_feat.addItem(feat)


    def selectFile(self):
        self.loadTableFile.setText(QFileDialog.getOpenFileName()[0])
        self.loadTable()
    
    def loadTable(self):
        """ Load features from a table file """
        filename = self.loadTableFile.text()
        self.mig.loadFeatureFile(filename)
        self.table.set_table()
        self.update_editfeature()
    

    def go_edit_feature(self):
        """ Edit an already loaded feature """
        featname = self.features_list.currentText()
        if not featname.startswith("Feat_"):
            featname = "Feat_"+featname
        featlist = self.mig.getFeaturesList()

        featimg = self.mig.image_feature_from_table( featname )
        edit_feat = EditFeature(  self.viewer, self.mig, self.labimg, self.table, featimg, featname, nb_values=2 )
        self.viewer.window.add_dock_widget( edit_feat, name="Edit "+featname )


class FeaturesTable(QWidget):
    """ Widget to visualize and interact with the measurement table """

    def __init__(self, napari_viewer, mig):
        super().__init__()
        self.viewer = napari_viewer
        self.mig = mig

        self.wid_table = QTableWidget()
        self.wid_table.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)
        self.setLayout(QGridLayout())
        self.layout().addWidget(self.wid_table)
        self.wid_table.clicked.connect(self.show_label)
        self.wid_table.setSortingEnabled(True)

        self.set_table()

    def show_label(self):
        """ When click on the table, show selected cell """
        if self.wid_table is not None:
            row = self.wid_table.currentRow()
            seglayer = self.viewer.layers["Cells"]
            seglayer.show_selected_label = False
            seglayer.visible = True
            headers = [self.wid_table.horizontalHeaderItem(ind).text() for ind in range(self.wid_table.columnCount()) ]
            labelind = None
            if "CellLabel" in headers:
                labelind = headers.index("CellLabel") 
            if labelind is not None and labelind >= 0:
                lab = int(self.wid_table.item(row, labelind).text())
                seglayer.selected_label = lab
                seglayer.show_selected_label = True
                #seglayer.refresh()


    def get_features_list(self):
        """ Return list of measured features """
        return [ self.wid_table.horizontalHeaderItem(ind).text() for ind in range(self.wid_table.columnCount()) ]

    def set_table(self, table=None, header=None):
        if table is None:
            table = self.mig.getFeaturesTable()
            header = self.mig.getFeaturesList()
        
        self.wid_table.clear()
        self.wid_table.setRowCount(len(table["CellLabel"]))
        self.wid_table.setColumnCount(len(table.keys()))

        for c, column in enumerate(table.keys()):
            column_name = column
            self.wid_table.setHorizontalHeaderItem(c, QTableWidgetItem(column_name))
            for r, value in enumerate(table.get(column)):
                item = QTableWidgetItem()
                if value == "":
                    value = "0"
                item.setData( Qt.EditRole, int(value))
                self.wid_table.setItem(r, c, item)

#class Projections(QWidget):
#    """ Widget to visualize projected chanels, can help to classify """

#    def __init__(self, napari_viewer):
#        self.viewer = napari_viewer

class EditFeature( QWidget ):
    """ GUI and shortcuts to edit a feature """

    def __init__( self, viewer, mig, labimg, table, featimg, featname, nb_values, channel=None ):
        """ Interface to edit a feature """
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.featname = featname
        self.nb_values = nb_values
        self.labimg = labimg
        self.channel = channel
        self.table = table

        layout = QVBoxLayout()
        lab = fwid.add_label( "Press 'i'/'d' to increase/decrease the value" )
        layout.addWidget( lab )

        ## slider
        value_line, self.new_value = fwid.slider_line( "Class value", 1, maxval=nb_values, step=1, value=1, show_value=True, slidefunc=None, descr="Choose the value to set clicked cells", div=1 )
        layout.addLayout( value_line )

        msg = fwid.add_label( "Right-click get the class value under the click \n [Control]+Left click set the clicked cell to the current class value" )
        layout.addWidget( msg )

        ## save feature image
        save_feat = fwid.add_button( "Export feature image", self.export_feature_image, descr="Save the image of cells colored by feature value" )
        add_class = fwid.add_button( "Add one class", self.add_one_class, descr="Add one possible class in the current feature")
        line = fwid.double_button( save_feat, add_class )
        layout.addLayout( line )

        ## finish the feature
        feat_done = fwid.add_button( str(self.featname)+" done", self.finish, descr="Close the edit interface for this feature", color=ut.get_color("done") )
        layout.addWidget( feat_done )

        self.setLayout( layout ) 
        self.add_one_feature( featimg, channel=channel )

    def finish( self ):
        """ Close this interface, close what should be closed """
        ut.remove_widget(self.viewer, "Edit "+self.featname)
        ut.remove_layer(self.viewer, self.featname+"Cells")
        if self.channel is not None:
            ut.remove_layer(self.viewer, "ProjC"+str(self.channel))
        if ut.has_widget(self.viewer, "Feature table"):
            self.table.set_table()

    
    def add_one_feature(self, featimg, channel=None):
        """ Launch editor of new feature """
        ## check that the nb of values is sufficient if the feature is already present
        featlist = self.mig.getFeaturesList()
        if (featlist is not None) and len(featlist)>0:
            if self.featname in featlist:
                maxval = self.mig.getFeatureMax( self.featname )
                if (maxval) > self.nb_values:
                    self.nb_values = maxval 
                    self.new_value.setMaximum( int(self.nb_values ) )

        ut.remove_layer(self.viewer, self.featname+"Cells")
        for layer in self.viewer.layers:
            layer.visible = False
        self.featlayer = self.viewer.add_labels(featimg, name=self.featname+"Cells", scale=(self.mig.scaleXY, self.mig.scaleXY), opacity=0.7, blending="additive")
        self.featlayer.contour = 5
        self.labimg.visible = True
        if (channel is not None) and ("ProjC"+str(channel) in self.viewer.layers):
            self.viewer.layers["ProjC"+str(channel)].visible = True
        
        self.active_bindings()

    def active_bindings( self ):
        """ Active the bindings for the layer """

        @self.featlayer.bind_key('i', overwrite=True)
        def increase_value(layer):
            self.new_value.setValue( int(self.new_value.value())%(self.nb_values) + 1 )
        
        @self.featlayer.bind_key('d', overwrite=True)
        def decrease_value(layer):
            self.new_value.setValue( (int(self.new_value.value()) - 2)%(self.nb_values) + 1 )
        
        @self.featlayer.mouse_drag_callbacks.append
        def click(layer, event):
            if event.type == "mouse_press":
                if event.button == 2:
                    # right click
                    pos = self.mig.get_coordinates(event.position)
                    value = self.featlayer.data[pos[0], pos[1]]
                    self.new_value.setValue( value )
                    return
                ## Set value to the cell under the cursor
                if (event.button == 1) and ("Control" in event.modifiers):
                    pos = self.mig.get_coordinates(event.position)
                    label = self.labimg.data[pos[0], pos[1]]
                    newvalue = int( self.new_value.value() )
                    self.setCellValue(label, newvalue)
                    return
        
    def setCellValue( self, label, newvalue ):
        """ Set the selected label to newvalue for the current feature """
        self.mig.change_cell_feature( self.featname, label, newvalue, self.featlayer.data )
        self.featlayer.refresh()
        
    def export_feature_image( self ):
        """ Save an image of the classified cells for the feature """
        self.mig.save_image( self.featlayer.data, endname="_feat_"+self.featname+".tif", imtype="uint8" ) 
        ## update the table with edited feature
        self.table.set_table()

    def add_one_class( self ):
        """ Add a new class to the feature """
        self.new_value.setMaximum( int(self.nb_values )  + 1 )
        self.nb_values = self.nb_values + 1
        ut.show_info("New class added to feature "+self.featname+" ("+str(self.nb_values)+" classes)")

class CreateFeature( QWidget ):
    """ GUI to create a new feature """
        
    def __init__(self, cc):
        """ Interface to create new feature """
        super().__init__()
        self.cc = cc
        #self.gNewFeat = QGroupBox("Create new feature")
        gnewfeat_layout = QVBoxLayout()
        ## name of the new feature
        self.feat_name = QLabel()
        self.feat_name.setText("")
        self.feat_name.setText("Feature name: "+self.cc.features_list.currentText())
        gnewfeat_layout.addWidget(self.feat_name)
        ## nb class of the new feature
        nvals = QHBoxLayout()
        nvals_lab = QLabel()
        nvals_lab.setText("Nb classes:")
        nvals.addWidget(nvals_lab)
        self.nbclass = QSpinBox()
        self.nbclass.setMinimum(2)
        self.nbclass.setMaximum(25)
        self.nbclass.setSingleStep(1)
        self.nbclass.setValue(2)
        nvals.addWidget(self.nbclass)
        gnewfeat_layout.addLayout(nvals)

        ## initialization method
        self.prefill = QComboBox()
        gnewfeat_layout.addWidget(self.prefill)
        self.prefill.addItem("Initialize all cells at 1")
        ## option to project and classify from that
        self.prefill.addItem("channel projection+threshold")
        self.create_projthreshold()
        gnewfeat_layout.addWidget(self.gProj)
        ## option to classify from touching the boundary or border
        self.prefill.addItem("Boundary cells")
        self.gui_boundary()
        gnewfeat_layout.addWidget(self.gBound)
        self.prefill.currentIndexChanged.connect(self.show_initoptions)

        ## show projection of a channel
        self.show_projection = QCheckBox(text="Show projected channel")
        gnewfeat_layout.addWidget(self.show_projection)
        self.show_projection.setChecked(True)
        self.show_projection.stateChanged.connect(self.show_initoptions)

        ## channel to project
        self.nchan = QHBoxLayout()
        self.nchan_lab = QLabel()
        self.nchan_lab.setText("channel to project:")
        self.nchan.addWidget(self.nchan_lab)
        self.nchannel = QSpinBox()
        self.nchannel.setMinimum(0)
        self.nchannel.setMaximum(self.cc.mig.nbchannels - 1)
        self.nchannel.setSingleStep(1)
        self.nchannel.setValue(0)
        self.nchan.addWidget(self.nchannel)
        gnewfeat_layout.addLayout(self.nchan)

        ## Button go to create new feature
        newfeat_go = QPushButton("Create new feature", parent=self)
        newfeat_go.setStyleSheet( 'QPushButton {background-color: #4a754d} ')
        gnewfeat_layout.addWidget(newfeat_go)
        self.show_initoptions()
        newfeat_go.clicked.connect(self.create_feature)

        self.setLayout(gnewfeat_layout)

    def set_name( self ):
        """ Update the name of the feature to create """
        self.feat_name.setText("Feature name: "+self.cc.features_list.currentText())


    def show_initoptions(self):
        """ Set visible parameters """
        self.gProj.setVisible( self.prefill.currentText() == "channel projection+threshold" )
        self.gBound.setVisible( self.prefill.currentText() == "Boundary cells" )
        show_nchan = (self.show_projection.isChecked()) or (self.prefill.currentText()=="channel projection+threshold") 
        self.nchannel.setVisible(show_nchan)
        self.nchan_lab.setVisible(show_nchan)
        if self.prefill.currentText() == "channel projection+threshold":
            text = "Initialize cells classification by mean projection of the selected channel \n"
            text += "Projected intensity is thresholded by: mean(intensites)*\'Threshold\' \n"
            text += "Cells are classified >0 if their proportion of positive pixels is greater than \'Prop area above threshold\' \n"
            ut.showOverlayText(self.cc.viewer, text)
        else:
            ut.removeOverlayText(self.cc.viewer)

    def gui_boundary( self ):
        """ GUI for boundary cells classification """
        self.gBound = QGroupBox( "Boundary classification" )
        gbound_layout = QVBoxLayout()

        ## option to get border cells or boundary cells
        self.border_cells = fwid.add_check( "Image border", False, None, descr="Class cells that are on image border as 2 or 3" ) 
        self.boundary_cells = fwid.add_check( "Tissue boundaries", True, None, descr="Class cells that are on tissue edges as 2 " ) 

        gbound_layout.addWidget( self.boundary_cells )
        gbound_layout.addWidget( self.border_cells )
        self.gBound.setLayout( gbound_layout )

    
    def create_projthreshold(self):
        """ GUI of projection initialization """
        self.gProj = QGroupBox("Projection+threshold")

        gproj_layout = QVBoxLayout()

        thres, self.thres_mean = fwid.value_line( "Threshold:", 1.5, descr="Threshold of mean projected intensity to consider a pixel positive" )
        gproj_layout.addLayout(thres)

        areaprop, self.area_prop = fwid.value_line( "Prop area above threshold:", 0.25, descr="Proportion of cell area that has to be positive (above threhsolde) to consider the cell as positive" ) 
        gproj_layout.addLayout(areaprop)

        self.gProj.setLayout(gproj_layout)
    
    def create_feature(self):
        """ Create a new feature classification and open edition widget """
        prefill_img = None
        feature_name = self.cc.features_list.currentText()
        if not feature_name.startswith("Feat_"):
            feature_name = "Feat_"+feature_name
            
        if feature_name+"Cells" in self.cc.viewer.layers:
            show_info("Feature "+feature_name+" already present, close it or choose another name")
            return
        
        ## Apply selected initialization  
        if self.prefill.currentText() == "Initialize all cells at 1":
            prefill_img = np.zeros(self.cc.mig.get_image_shape(in2d=True), np.uint8)
            prefill_img[self.cc.labimg.data>0] = 1
            self.cc.mig.fillFeature( feature_name, 1 )

        if self.prefill.currentText() == "channel projection+threshold":
            prefill_img = self.project_and_threshold(feature_name)

        if self.prefill.currentText() == "Boundary cells":
            prefill_img = self.class_boundary_cells( feature_name )
        
        channel = int(self.nchannel.text())
        if self.show_projection.isChecked():
            ## Show projection image if option is selected
            if "ProjC"+str(channel) not in self.cc.viewer.layers:
                img = self.cc.mig.get_channel(channel)
                proj = np.mean(img, axis=0)
                colmap = self.cc.viewer.layers["originalChannel"+str(channel)].colormap
                projlay = self.cc.viewer.add_image(proj, name="ProjC"+str(channel), scale=(self.cc.mig.scaleXY, self.cc.mig.scaleXY), colormap=colmap, blending="additive")
                projlay.contrast_limits = ut.quantiles(proj) 
                projlay.gamme = 0.95
            
        nb_values = int(self.nbclass.text())
        channel = int(self.nchannel.text())
        edit_feat = EditFeature(  self.cc.viewer, self.cc.mig, self.cc.labimg, self.cc.table, prefill_img, feature_name, nb_values, channel )
        self.cc.viewer.window.add_dock_widget( edit_feat, name="Edit "+feature_name )
        #self.cc.add_one_feature(feature_name, prefill_img, nb_values, channel)
        self.cc.features_list.addItem( feature_name )
        self.hide()
    
    def project_and_threshold(self, featname):
        """ Calculate the projection and initialize classification by thresholding it """
        chan = int(self.nchannel.text())
        threshold_mean = float(self.thres_mean.text())
        threshold_fill = float(self.area_prop.text()) 
        return self.cc.mig.classifyCells( featname, chan, threshold_mean, threshold_fill )

    def class_boundary_cells( self, featname ):
        """ Class the cells according to touching boundary/edge or not """
        boundary = int(self.boundary_cells.isChecked() * 2) 
        border = int(self.border_cells.isChecked() * 2 + boundary / 2)
        return self.cc.mig.classifyBoundaryCells( featname, border, boundary ) 

class GetScales( QWidget ):
    """ Get scales information if load directly from classify """

    def __init__( self, viewer, mig, classify_cells ):
        """ Interface to get the metadata """
        super().__init__()
        self.viewer = viewer
        self.mig = mig
        self.classify_cells = classify_cells

        layout = QVBoxLayout()
        ## scalexy parameter
        scalexy_line, self.scale_xy = fwid.value_line( "Scale XY:", self.mig.scaleXY, descr="Pixel size in the lateral dimensions (XY), in microns" )
        layout.addLayout(scalexy_line)
        ## scalez parameter
        scalez_line, self.scale_z = fwid.value_line( "Scale Z:", self.mig.scaleZ, descr="Pixel size in the axial dimension (Z), in microns" )
        layout.addLayout(scalez_line)
        ## load filename
        load_line, self.segmented_cells = fwid.file_line( "Segmented cells file:", self.mig.junction_filename(dim=2,ifexist=True), "Select the segmented cells file to load", descr="File containing the segmented cells, should be a tif file with integer values" )
        layout.addLayout( load_line )

        ## btn go
        btn_update = fwid.add_button( "Update", self.update_scales, descr="Update the scale of the image and load the segmented cells", color=ut.get_color("go") )
        layout.addWidget(btn_update)
        self.setLayout(layout)

    def update_scales( self ):
        """ Go update scale """
        self.mig.scaleXY = float( self.scale_xy.text() )
        self.mig.scaleZ = float( self.scale_z.text() )
        for chan in range(self.mig.nbchannels):
            self.viewer.layers['originalChannel'+str(chan)].scale = [self.mig.scaleZ, self.mig.scaleXY, self.mig.scaleXY]
        ut.remove_all_widget(self.viewer)
        self.mig.load_segmentation( self.segmented_cells.text() )
        done = self.mig.popFromJunctions()
        if done < 0:
            return
        self.classify_cells(self.mig, self.viewer)
    


import numpy as np
import cv2
import os
import random
import scipy.ndimage as ndimage
from skimage.measure import regionprops, label
from napari.utils.notifications import show_info
import fish_feats.Utils as ut
from fish_feats.Association import associateNucleus, associateNucleusOverlap
from skimage.filters import sato
from skimage.morphology import skeletonize, binary_closing
from skimage.segmentation import find_boundaries
from skimage.segmentation import clear_border
from skimage.measure import label
from skimage.exposure import adjust_gamma
import time
import tempfile

"""
    Functions to segment junctions or nuclei
"""

def run_epyseg(input_folder):
    import tensorflow as tf

    # libraries loaded checking epyseg to see if everything is functional
    try:
        from epyseg.img import Img
        from epyseg.deeplearning.deepl import EZDeepLearning
        from epyseg.deeplearning.augmentation.meta import MetaAugmenter
        from epyseg.deeplearning.augmentation.generators.data import DataGenerator
        deepTA = EZDeepLearning()
    except:
        print('EPySeg failed to load.')
    
    # Load a pre-trained model
    pretrained_model_name = 'Linknet-vgg16-sigmoid-v2'
    pretrained_model_parameters = deepTA.pretrained_models[pretrained_model_name]

    deepTA.load_or_build(model=None, model_weights=None,
                             architecture=pretrained_model_parameters['architecture'], backbone=pretrained_model_parameters['backbone'],
                             activation=pretrained_model_parameters['activation'], classes=pretrained_model_parameters['classes'],
                             input_width=pretrained_model_parameters['input_width'], input_height=pretrained_model_parameters['input_height'],
                             input_channels=pretrained_model_parameters['input_channels'],pretraining=pretrained_model_name)
    #epydir = os.path.join(os.path.abspath(".."), "epyseg_net")
    #deepTA.load_weights( os.path.join(epydir,'Linknet-vgg16-sigmoid-v2.h5') )

    input_val_width = 256
    input_val_height = 256
    
    input_shape = deepTA.get_inputs_shape()
    output_shape = deepTA.get_outputs_shape()
    if input_shape[0][-2] is not None:
        input_val_width=input_shape[0][-2]
    if input_shape[0][-3] is not None:
        input_val_height=input_shape[0][-3]
    #print(input_shape)
    deepTA.compile(optimizer='adam', loss='bce_jaccard_loss', metrics=['iou_score'])
      
    range_input = [0,1]
    input_normalization = {'method': 'Rescaling (min-max normalization)',
                        'individual_channels': True, 'range': range_input, 'clip': True} 

    predict_generator = deepTA.get_predict_generator(
            inputs=[input_folder], input_shape=input_shape,
            output_shape=output_shape, 
            default_input_tile_width=input_val_width, 
            default_input_tile_height=input_val_height,
            tile_width_overlap=32,
            tile_height_overlap=32, 
            input_normalization=input_normalization, 
            clip_by_frequency={'lower_cutoff': None, 'upper_cutoff': None, 'channel_mode': True} )

    post_process_parameters={}
    post_process_parameters['filter'] = None
    post_process_parameters['correction_factor'] = 1
    post_process_parameters['restore_safe_cells'] = False ## no eff
    post_process_parameters['cutoff_cell_fusion'] = None
    post_proc_method = 'Rescaling (min-max normalization)'
    post_process_parameters['post_process_algorithm'] = post_proc_method
    post_process_parameters['threshold'] = None  # None means autothrehsold # maybe add more options some day

    predict_output_folder = os.path.join(input_folder, 'predict')
    print("Starting segmentation with EpySeg.....")
    deepTA.predict(predict_generator, 
                output_shape, 
                predict_output_folder=predict_output_folder,
                batch_size=1, **post_process_parameters)
    
    deepTA.clear_mem()
    if not os.access(predict_output_folder, os.W_OK):
        os.chmod(predict_output_folder, stat.S_IWUSR)
    #deepTA = None
    del deepTA

def run_epyseg_onimage(img, filedir, filename, verbose=True):
    from PIL import Image
    import os

    binimg = None
    tmpdir_path = None
    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            print("tmp dir "+str(tmpdir))

            ### empty directory if exists
            inputname = filename+"_junctions.tif"
            with Image.fromarray(img) as im:
                im.save(os.path.join(tmpdir,inputname))
            try:
                predict_output_folder = os.path.join(tmpdir, 'predict')
                os.makedirs(predict_output_folder, exist_ok=True)
            except:
                print("Warning, issue in creating "+predict_output_folder+" folder")
    
            ## run Epyseg on tmp directory (contains current image)
            run_epyseg(tmpdir)
    
            ## return result and delete files
            im = Image.open(os.path.join(tmpdir,"predict",inputname)) 
            binimg = np.copy(im)
            os.chmod(os.path.join(tmpdir, "predict", inputname), 0o777)
            im.close()
            os.remove( os.path.join(tmpdir, "predict", inputname) )
    except BaseException as e:
        print("Error in running EpySeg: "+str(e))
        print("Warning, folder not deleted")
        tmpdir_path = tmpdir
        pass

    return junctions_to_label(binimg), tmpdir_path

def run_epyseg_onimage_fold(img, filedir, filename, verbose=True):
    from PIL import Image
    import os
    import shutil
    import stat
    
     ## prepare data
    tmpdir = os.path.join(filedir,"TmpEpySegDir_"+str(random.randint(0,5000)))

    ### empty directory if exists
    if os.path.exists(tmpdir):
        os.chmod(tmpdir, stat.S_IRUSR|stat.S_IRGRP|stat.S_IROTH|stat.S_IWUSR|stat.S_IWGRP|stat.S_IWOTH|stat.S_IXUSR)
        shutil.rmtree(tmpdir, ignore_errors=True)
    os.mkdir(tmpdir)
    inputname = filename+"_junctions.tif"
    im = Image.fromarray(img)
    im.save(os.path.join(tmpdir,inputname))
    try:
        predict_output_folder = os.path.join(tmpdir, 'predict')
        os.makedirs(predict_output_folder, exist_ok=True)
    except:
        print("Warning, issue in creating "+predict_output_folder+" folder")
    
    ## run Epyseg on tmp directory (contains current image)
    run_epyseg(tmpdir)
    
    ## return result and delete files
    im = Image.open(os.path.join(tmpdir,"predict",inputname))
    try:
        os.chmod(predict_output_folder, stat.S_IRUSR|stat.S_IRGRP|stat.S_IROTH|stat.S_IWUSR|stat.S_IWGRP|stat.S_IWOTH|stat.S_IXUSR)
        shutil.rmtree(predict_output_folder, ignore_errors=True)
        os.chmod(tmpdir, stat.S_IRUSR|stat.S_IRGRP|stat.S_IROTH|stat.S_IWUSR|stat.S_IWGRP|stat.S_IWOTH|stat.S_IXUSR)
        shutil.rmtree(tmpdir, ignore_errors=True)
    except:
        print("Folder "+tmpdir+" not deleted, you can delete it manually")

    return junctions_to_label(im)

def junctions_to_label(img):
    """ convert junctions result to labels map """
    return( label(np.invert(img), background=0, connectivity=1) )

def run_cellpose(img, scaleXY, diameter=7, verbose=True):
    from cellpose import models
    model = models.CellposeModel(gpu=True, model_type='cyto3') 
    diamet = diameter/scaleXY   ## increase it ?
    mask, flow, style = model.eval(img, invert=False, diameter=diamet, channels=[0,0], do_3D=False, cellprob_threshold=0.05)
    ## convert cellpose result to label image (cellpose result are not touching border, make it as junctions)
    return fromcellpose_tojunctions(mask)

def fromcellpose_tojunctions(mask):
    bined = binary_closing( find_boundaries(mask), footprint=np.ones((10,10)) )
    return junctions_to_label( skeletonize( bined ) )

def segmentJunctions(img, mode, scaleXY, imagedir, imagename, diameter=7, chunking=1000, verbose=True):
    """ segment junctions from ZO1 signal, either with EpySeg or CellPose """
    if (mode == 0) or (isinstance(mode, str) and mode.lower() == "epyseg"):
        show_info("Starting segmentation with Epyseg...")
        seged, tmpdir_path = run_epyseg_onimage(img, imagedir, imagename, verbose)
        if tmpdir_path is not None:
            if os.path.exists(tmpdir_path):
                os.chmod(tmpdir_path, stat.S_IRUSR|stat.S_IRGRP|stat.S_IROTH|stat.S_IWUSR|stat.S_IWGRP|stat.S_IWOTH|stat.S_IXUSR)
                shutil.rmtree(tmpdir_path, ignore_errors=True)
    if (isinstance(mode, str) and mode.lower() == "cellpose"):
        show_info("Starting segmentation with CellPose...")
        seged = run_cellpose(img, scaleXY, diameter=diameter, verbose=verbose)
    if (isinstance(mode, str) and mode.lower() == "epyseg-dask"):
        from fish_feats.DaskedEpyseg import run_dasked_epyseg
        show_info("Starting segmentation with dasked Epyseg...")
        overlap = diameter/scaleXY*0.25
        tmp_dir = tempfile.TemporaryDirectory()
        tmp_file = os.path.join( tmp_dir.name, "tmp.zarr" ),
        dask_seg = run_dasked_epyseg( img, imagedir, imagename, overlap=overlap, iou_depth=2, iou_threshold=0.1, chunksize=(chunking, chunking), tmpfile=tmp_file )
        seged = dask_seg.compute()
    return clear_border(seged)

def prepJunctions( img, do_clahe=True, clahe_grid=10 ):
    img = cv2.normalize(src=img, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    if do_clahe:
        clahe = cv2.createCLAHE( clipLimit=2.0, tileGridSize=(clahe_grid, clahe_grid) )
        img = clahe.apply(img)
    return img

def prepJunctions_sharp(img, medBlur, sharp):
    #from skimage import restoration
    junc = np.max(img, axis=0)
    junc = cv2.normalize(src=junc, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    junc = cv2.medianBlur(junc,medBlur)
    #rad = 2.0/scaleXY
    #juncbg = restoration.rolling_ball(junc, radius=rad)
    #res = junc-juncbg
    res = junc
    #res[res<0] = 0
    #res = cv2.normalize(src=res, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    #b1 = cv2.GaussianBlur(res, (0,0), 20,20,0)
    #b2 = cv2.GaussianBlur(res, (0,0), 1,1,0)
    #res = b1-b2
    #res[res<0] = 0
    #res = cv2.fastNlMeansDenoising(res,None, 7, 21, 7)
    #res = cv2.bilateralFilter(res, 9, 31,31)
    sharpen_kernel = np.array( [[-1,-1,-1],[-1,sharp,-1],[-1,-1,-1]])
    res = cv2.filter2D(res, -1, sharpen_kernel)
    #res = cv2.normalize(src=res, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
    return res

def fill_posmax(array, meanimg, meanmax):
    for z in range(meanimg.shape[0]):
        array += z * (meanimg[z,]==meanmax) * (array==0)
    return array

def uniform_filtering( array, smooth ):
    return ndimage.uniform_filter( array, size=(smooth, smooth) )

def maximum_filtering( array, sizexy ):
    filted = ndimage.maximum_filter(array, size=(1, sizexy, sizexy))
    return filted

def local_max_proj_parallel(img, largexy=40, smooth=2):
    """ Apply local max in parallel """
    from skimage.util import apply_parallel
    meanimg = ndimage.maximum_filter(img, size=(1, largexy, largexy))
    meanmax = np.max(meanimg, axis=0)
    posmax = np.zeros(meanimg.shape[1:3])
    for z in range(meanimg.shape[0]):
        posmax += z * (meanimg[z,]==meanmax) * (posmax==0)
    
    posmax = apply_parallel( uniform_filtering, posmax, depth=10, extra_arguments=(int(largexy*1.5),), compute=True )
    
    projimg = np.zeros(posmax.shape)
    for z in range(img.shape[0]):
        projimg += img[z,] / (1+np.abs(posmax-z)) * (np.abs(posmax-z)<=3)/7

    return apply_parallel( uniform_filtering, projimg, depth=10, extra_arguments=(smooth,), compute=True )

def local_max_proj(img, largexy=40, smooth=2):
    """ Do local max projection of junctions """

    ## get position of local maxima
    meanimg = ndimage.maximum_filter(img, size=(1, largexy, largexy))
    posmax = np.argmax(meanimg, axis=0)
    posmax = ndimage.uniform_filter(posmax, size=(int(largexy*1.5), int(largexy*1.5)))
    
    ## project the intensity values around the local max
    projimg = np.zeros(posmax.shape)
    nz = 2
    for dz in range(-nz, nz):
        curpos = np.uint8(posmax+dz)
        curpos[curpos<0] = 0
        curpos[curpos >= img.shape[0]] = img.shape[0]-1
        projimg += np.take_along_axis(img, curpos[np.newaxis], axis=0)[0]/(1+abs(dz)) * 0.1

    #for z, zimg in enumerate(img):
    #    dist2max = np.abs(posmax-z)
    #    within = (dist2max <= 3)
    #    if np.sum(within) > 0:
    #        projimg += zimg / (1+dist2max) * (within)/7

    projimg = ndimage.uniform_filter(projimg, size=(smooth, smooth))
    return projimg


def image_tubeness(img):
    img = sato(img, sigmas=range(1, 10, 2), black_ridges=False, mode='reflect', cval=0)
    return ndimage.uniform_filter(img, size=(10,10)) 


def normalizeQuantile(img, qmin=0.001, qmax=0.99, vmax=255):
    quants = np.quantile(img, [qmin, qmax])
    img = (img-quants[0])/(quants[1]-quants[0])*vmax
    img = np.clip(img, 0, quants[1])
    return img

def gaussBlur(im, rad):
    return cv2.GaussianBlur(im, (0,0), rad, rad, 0)


def preprocessNucleiMedianFilter(img, radxy, radz):
    return ndimage.median_filter(img, size=(radz, radxy, radxy) )

def preprocessRemoveBg(img, radxy, radz):
    bg = ndimage.uniform_filter(img, size=(radz, radxy, radxy) )
    img = img - bg
    img[img<0] = 0
    return img

def preprocessRemoveBg2D(img, radxy):
    bg = ndimage.uniform_filter(img, size=(radxy, radxy) )
    img = np.float32(img)
    img = img - bg
    img[img<0] = 0
    img = np.uint8(img)
    return img

def prepNuclei(img):
    """ Normalize by quantiles """
    #img[img<0] = 0
    img = adjust_gamma(img, 1.05)
    quants  = np.quantile(img, [0.001, 0.99])
    img = np.clip(img, quants[0], quants[1])
    img = (img-quants[0])/(quants[1]-quants[0])
    return img

def stardist2D(img, prob, over):
    """ run stardist model, segment nuclei in 2D """
    try:
        from csbdeep.utils import Path, normalize
        from stardist.models import StarDist2D
    except ModuleNotFoundError:
        ut.show_error("Stardist2D not installed, please install stardist with pip install stardist")
        return None
    
    model = StarDist2D.from_pretrained('2D_paper_dsb2018')  ## 2D_versatile_fluo, find more nuclei than paper_dsb2018
    axis_norm = None ## normalize jointly
    labs = []
    tiling = (int(img.shape[1]/2000)+1, int(img.shape[2]/2000)+1)
    nuclei = np.zeros(img.shape, "uint16")
    for ind, im in enumerate(img):
        #if np.mean(im)<0.1:    ## nearly empty slice, makes it bug
        #    labels = np.zeros(im.shape)
        #else:
        #    im = normalize(im, 1, 99.8, axis=axis_norm)
        labels, details = model.predict_instances(im, n_tiles=tiling, prob_thresh=prob, nms_thresh=over, show_tile_progress=True)
        nuclei[ind,] = labels
    return nuclei

def getNuclei_stardist2DAsso3D(nucimg, scaleXY, proba=0.55, overlap=0.1, assoMode="Munkres", assolim=3, threshold_overlap=0.25, verbose=True):
    """ Segment nuclei with Stardist2D and reconstruct in 3D - return the nuclei list """
    
    ## segment 2D
    labnuc = stardist2D(nucimg, prob=proba, over=overlap)
    if labnuc is None:
        return None
    ## reconstruct 3D
    if assoMode == "Munkres":
        labels = associateNucleus(labnuc, dlimit=assolim, scaleXY=scaleXY)  ## distance 2D in micrcons  -2.5
    if assoMode == "Overlap":
        labels = associateNucleusOverlap(labnuc, threshold_overlap=threshold_overlap)
    return finishNuclei(labels, 2, False, verbose)


def finishNuclei( nuclab, minz=2, convexify=False, verbose=True ):
    start_time = time.time()
    regions = regionprops( nuclab )
    for r in regions:
        ## if only 2 or 3 pixels, too small, remove
        if r.area < 4:
            nuclab[nuclab==r.label] = 0
            continue 
        ## in less than minz slices, too small, remove
        if (r.bbox[3]-r.bbox[0]) < minz:
            nuclab[nuclab==r.label] = 0
        else:
            if convexify:
                nuclab[r.bbox[0]:r.bbox[3], r.bbox[1]:r.bbox[4], r.bbox[2]:r.bbox[5]][r.image_convex] = r.label

    if verbose:
        print("Nb nuclei found: "+str(len(np.unique(nuclab))))
        print("Post-processing nuclei took {:.3f}".format((time.time()-start_time)/60)+" min")
    return nuclab
    

def initialize_cellpose():
    from cellpose import models 
    model = models.CellposeModel(gpu=True, model_type='nuclei') 
    return model

def run_cellpose_nuclei(model, nucimg, norm, diameter, scaleXY, scaleZ, threshold, flow_threshold, resample=True, in3D=True, stitch_threshold=0.25, verbose=True):
    diamet = diameter/scaleXY
    if verbose:
        print("Starting 3D nuclei segmentation with CellPose3D...")
    if in3D:
        anisotropy = scaleZ/scaleXY
    else:
        anisotropy = 1.0
    mask, flow, style = model.eval(nucimg, invert=False, normalize=norm, diameter=diamet, channels=[0,0], channel_axis=0, z_axis=1, resample=resample, do_3D=in3D, stitch_threshold=stitch_threshold, anisotropy=anisotropy, flow_threshold=flow_threshold, cellprob_threshold=threshold)
    if verbose:
        print("3D nuclei segmentation done")
    return mask

def run_cellpose_nuclei_dask( model, nucimg, norm, diameter, scaleXY, scaleZ, threshold, flow_threshold, resample=True, in3D=True, stitch_threshold=0.25, chunk=1000, verbose=True ):
    from fish_feats.cellpose_dask import dask_segment
    diamet = diameter/scaleXY
    if verbose:
        print("Starting 3D nuclei segmentation with CellPose3D...")
    if in3D:
        anisotropy = scaleZ/scaleXY
    else:
        anisotropy = 1.0
    
    tmp_dir = tempfile.TemporaryDirectory()
    nucimg = np.moveaxis( nucimg, 0, 3 )
    dasky = dask_segment(
        nucimg, channels=[0,0], 
        diameter = [diamet, diamet], fast_mode=True,
        use_anisotropy=in3D, iou_depth = 10, 
        iou_threshold=0.4,
        chunksize=(chunk, chunk, chunk, 1),
        dim = 4,
        do_3D=in3D,
        stitch_threshold = stitch_threshold,
        resample = resample,
        cellprob_threshold = threshold,
        flow_threshold = flow_threshold,
        model = model,
        z_axis=0,
        tmp_file = os.path.join( tmp_dir.name, "tmp.zarr" ),
    )
    print(dasky)

    mask = dasky.compute()
    print("done")
    if verbose:
        print("3D nuclei segmentation done")
    return mask


def getNuclei_cellpose3D(model, nucimg, diameter, scaleXY, scaleZ, threshold=0.1, flow_threshold=0.4, resample=True, in3D=True, stitch_threshold=0.25, dask=False, chunk=1000, verbose=True):
    ## pretreat
    if not dask:
        mask = run_cellpose_nuclei(model, nucimg, False, diameter, scaleXY, scaleZ, threshold, flow_threshold, resample, in3D, stitch_threshold, verbose)
    else:
        mask = run_cellpose_nuclei_dask(model, nucimg, False, diameter, scaleXY, scaleZ, threshold, flow_threshold, resample, in3D, stitch_threshold, chunk=chunk, verbose=verbose)
    return finishNuclei( mask, 2, convexify=False, verbose=verbose )

def prepElectroporation(img, radxy, radz):
    from skimage import exposure
    img = exposure.adjust_gamma(img, 0.9)
    imin = np.quantile(img, 0.001)
    imax = np.quantile(img, 0.999)
    img = (img-imin)/(imax-imin)*255
    img[img<0] = 0
    img = np.uint8(img)
    img = ndimage.median_filter(img, size=(radxy, radxy, radz) )
    #img = ndimage.gaussian_filter(img, (radxy,radxy,radz) )
    return img

def getElectroporated(img):
    bins = []
    msig = np.mean(img)*0.95#+0.5*np.std(img)  ## signal is sparse, so mean is low
    #print(str(np.mean(img))+" "+str(np.std(img)))
    for im in img:
        #print(np.mean(im))
        if np.mean(im) > msig:   ## contains some signal
            th, bim = cv2.threshold(im, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)
        else:
            bim = np.zeros(im.shape)
        bins.append(bim)
    bins = np.array(bins)
    s = ndimage.generate_binary_structure(3,3)
    labels, num = ndimage.label(bins, structure=s)
    vols = ndimage.sum(bins, labels, range(num))
    centers = ndimage.center_of_mass(bins, labels, range(num))
    return labels, vols, centers
